/* glmrunner.js
GLM computations 
For BrainVoyager QX 2.1 and higher
HB BI 2012
*/

// widget names
var btnLoadFilenames, lblVMRname, btnLoadVMR, chbCorrectSerCor, chbVTC, chbFMR, spbNrGLMs;
var chbMultiSubjGLM, chbRFXGLM, chbSeparateStudies, chbSeparateSubjects, rbtnZtrfBaseline, rbtnZtrf, rbtnPSC;
var rbtnReadTextfile, rbtnSelectViaDlg, btnStart;

// 
const TEXTFILE_FOR_SSGLM        = 1;
const TEXTFILE_FOR_POST_SSGLM   = 2;
//var projfilenames; 
const FILETYPE_UNKNOWN          = 0;
const FILETYPE_FMR              = 1;
const FILETYPE_VTC              = 2;

var scriptObj = new Object; 
scriptObj.scriptNameForUser     = "glmrunner"; 
scriptObj.dialogFileName        = "glmrunner_v01.ui"; 
scriptObj.dialogAccessName      = "glmDlg"; 

var projfilenames; 

//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

scriptObj.startProcessing = function() {

 //   var success = this.getDesMatAndFuncFiles();
    var success = (projfilenames.length > 0);
    if (success) {
        Compute_GLMs(); 
    }
    this.glmDlg.close();
}

function Compute_GLMs() {
    
	var counter, doc, funcfile, desmatfile;
    
    // always run single-subject
    var nroffiles = (projfilenames.length);
    var filetype = (chbVTC.checked ? FILETYPE_VTC : FILETYPE_FMR);
    var anatname = lblVMRname.text;
    if ((anatname.length < 1) && FILETYPE_VTC) BrainVoyagerQX.PrintToLog("Please note: this option requires an anatomical file.");
    
    BrainVoyagerQX.PrintToLog("Start calculating " + nroffiles + " single-study GLMs..."); 
    for (counter = 0; counter < nroffiles; counter++) {	 
    
        funcfile = projfilenames[counter][0];
        BrainVoyagerQX.PrintToLog("Functional file: " + funcfile);                   
        desmatfile = projfilenames[counter][1];                   
            
        if (filetype == FILETYPE_FMR) {		
                BrainVoyagerQX.PrintToLog("Opening " + funcfile + "..."); // functional data (*.fmr + *.stc)
                doc = BrainVoyagerQX.OpenDocument(funcfile); 
        } else if (filetype == FILETYPE_VTC){
                BrainVoyagerQX.PrintToLog("Opening " + anatname + "..."); // anatomical data (*.vmr)
                doc = BrainVoyagerQX.OpenDocument(anatname); 
                BrainVoyagerQX.PrintToLog("Linking " + funcfile + "..."); // functional data (*.vtc)
                doc.LinkVTC(funcfile); 
        }

        BrainVoyagerQX.PrintToLog("Loading " + desmatfile + "...");
        doc.LoadSingleStudyGLMDesignMatrix(desmatfile);	// design matrix (*.sdm)
        doc.CorrectForSerialCorrelations = chbCorrectSerCor.checked;
        BrainVoyagerQX.PrintToLog("Correct serial correlations: " + (doc.CorrectForSerialCorrelations ? "yes" : "no"));
        doc.ComputeSingleStudyGLM();
        
        var glmname = removeextension(funcfile) + ".glm";		
        BrainVoyagerQX.PrintToLog("Saving " + glmname + "...");	
        doc.SaveGLM(glmname); 
        doc.ShowGLM(); 
        doc.Close();
    } // end counter		  
	
	if  ((chbMultiSubjGLM.checked) || (chbRFXGLM.checked)) {
		
         var anatname = lblVMRname.text;
         var mdmname = Generate_MDM_file(projfilenames, anatname); // In BrainVoyager QX 2.1 only for VTCs 
        
		 BrainVoyagerQX.PrintToLog("Opening " + anatname + "..."); // anatomical data (*.vmr)
		 doc = BrainVoyagerQX.OpenDocument(anatname); 
		 doc.LoadMultiStudyGLMDefinitionFile(mdmname);
		 doc.PSCTransformStudies 			= rbtnPSC.checked;
		 doc.ZTransformStudies 				= rbtnZtrf.checked;
		 doc.ZTransformStudiesBaselineOnly 	= rbtnZtrfBaseline.checked;
		 doc.SeparationOfSubjectPredictors 	= chbSeparateSubjects.checked;
		 doc.SeparationOfStudyPredictors 	= chbSeparateStudies.checked;
		 BrainVoyagerQX.PrintToLog("Predictor separation: " + (doc.SeparationOfSubjectPredictors ? "per subject" : (doc.SeparationOfStudyPredictors ? "per run" : "not set")));
		 BrainVoyagerQX.PrintToLog("Data normalization: " + (doc.ZTransformStudies ? "z-transform" : (doc.ZTransformStudiesBaselineOnly ? "z-transform (baseline)" : 
										(doc.PSCTransformStudies ? "percent signal change " : "not set"))));
		 
         if (chbMultiSubjGLM.checked) {
            doc.ComputeMultiStudyGLM();
            var glmname = removeextension(mdmname) + ".glm";		
            BrainVoyagerQX.PrintToLog("Saving " + glmname + "...");	
            doc.SaveGLM(glmname);
            doc.ShowGLM();
         }
         if (chbRFXGLM.checked) {        
            doc.ComputeRFXGLM();
            var rfxglmname = removeextension(mdmname) + "_rfx.glm";		
            BrainVoyagerQX.PrintToLog("Saving " + rfxglmname + "...");	
            doc.SaveGLM(rfxglmname);
            doc.ShowGLM();
         }
	}
}

function Generate_MDM_file(projfilenames, anatname) {
    
    var mdmname;
    BrainVoyagerQX.ShowLogTab();
	BrainVoyagerQX.PrintToLog("Start creating MDM...\nOpening anatomical file: " + anatname);
    var nroffiles = projfilenames.length;
    var anat = BrainVoyagerQX.OpenDocument(anatname);
	if (anat == undefined) {
		BrainVoyagerQX.PrintToLog("Invalid anatomical filename.");
	} else {
        anat.ClearMultiStudyGLMDefinition();
        for (counter = 0; counter < nroffiles; counter++) {
            BrainVoyagerQX.PrintToLog("Adding:\n" + projfilenames[counter][0] + "\n" + projfilenames[counter][1]);
            anat.AddStudyAndDesignMatrix(projfilenames[counter][0], projfilenames[counter][1]);
        }
		mdmname = removeextension(projfilenames[0][1]) + "_" + nroffiles + "studies.mdm";// modify *.sdm name
		BrainVoyagerQX.PrintToLog("Name for MDM file: " + mdmname);
		anat.SaveMultiStudyGLMDefinitionFile(mdmname);
    }
    return mdmname;
}

function removeextension(filename) {
	var extlength = getextensionlength(filename);
	var name = "";
	for (i=0; i<(filename.length-extlength); i++) {
		letter = filename[i];
		name += letter;
	}	
	return name;
}

function getextensionlength(filename) {
	var extensionlength, letter, dotpos;
	for (i=filename.length; i>0; i--) {
		letter = filename[i];
		if (letter == ".") {
			dotpos = i;					
			break;
		}	
	}
	extensionlength = filename.length - dotpos;
	return extensionlength;
}

scriptObj.getFile = function(filetype) {

    var success = true;
    BrainVoyagerQX.SetPluginStringParameter("FileType", filetype);  
    BrainVoyagerQX.SetPluginStringParameter("Command", "GetFileName"); 
    BrainVoyagerQX.ExecutePlugin(); 
    var filename = BrainVoyagerQX.GetPluginStringParameter("Filename");
    return filename;
}

scriptObj.getAnatFile = function() {

	BrainVoyagerQX.PrintToLog("Get anatomical file (*.vmr)..."); 
    var filetype = "Anatomical file (*.vmr)";
    var filename = this.getFile(filetype);
    if (filename.length > 0) {
        lblVMRname.text = filename;
        BrainVoyagerQX.PrintToLog("Name: " + filename);
    }
}

scriptObj.getDesMatAndFuncFiles = function() {

    var success = true;
	BrainVoyagerQX.PrintToLog("Get files..."); 
    var filetype = "*.*";
    if (chbFMR.checked) {
        filetype = "*.fmr";
    } else {
        filetype = "*.vtc";
    } 
    
    // select files manually
    projfilenames = new Array();
    if (rbtnSelectViaDlg.checked == 1) {
    
        var nroffiles = spbNrGLMs.value;
        if (nroffiles < 1) BrainVoyagerQX.PrintToLog("Please note: number of selected files is 0.");
        BrainVoyagerQX.PrintToLog("Start selecting: " + nroffiles + " files..."); 
        var filecounter;
        // loop over indicated number of files
        for (filecounter = 0; filecounter < nroffiles; filecounter++) {
            
            var groupnames = new Array(2);
            var funcfilename = this.getFile(filetype);
            if (funcfilename.length > 0) groupnames[0] = funcfilename;     
            var desmatname = this.getFile("*.sdm");
            if (desmatname.length > 0) groupnames[1] = desmatname;
            projfilenames.push(groupnames);
            
        }
        this.writeToTextFile(projfilenames, TEXTFILE_FOR_SSGLM, "glm");// stage, targetfilenames
        BrainVoyagerQX.PrintToLog("Files selected. If all parameters have been set, the Start button can be clicked.");
        
    // read filenames from textfile    
    } else if (rbtnReadTextfile.checked == 1) {
    
        BrainVoyagerQX.SetPluginStringParameter("Command", "GetTextFileName");
        BrainVoyagerQX.PrintToLog("Select text file with source filenames..."); 
        BrainVoyagerQX.ExecutePlugin();
        if (this.noErrors()) {
            var textfilename = BrainVoyagerQX.GetPluginStringParameter("Filename");
            BrainVoyagerQX.PrintToLog("Selected text filename: " + textfilename);
            this.readTextFileIntoProjectNamesArray(textfilename, TEXTFILE_FOR_SSGLM); 
        } else {
            success = false;
        }
    }    
    return success;
}

scriptObj.writeToTextFile = function(projfilenames, kindoffiles, stage) {//, stage, targetfilenames) 
    
    var filecounter;
    var success;
    
    // write the textfile to the path of the first file
    var filetype = projfilenames[0][0].substring((projfilenames[0][0].length-3), projfilenames[0][0].length);
    var name = String(this.getPath(projfilenames[0][0]) + "filenames_" + stage + "_" + filetype.toUpperCase() + ".txt");
    var filenamesfile = new QFile(name);     
    filenamesfile.open(new QIODevice.OpenMode(QIODevice.WriteOnly)); 
    var textstr = new QTextStream(filenamesfile); 
    textstr.writeDouble(projfilenames.length);
    textstr.writeString("\n");
    if (kindoffiles == TEXTFILE_FOR_SSGLM) {
        BrainVoyagerQX.PrintToLog("Write design matrix and functional names...");
        for (filecounter = 0; filecounter < projfilenames.length; filecounter++) {
            textstr.writeString(projfilenames[filecounter][0] + "\n"); // source file (VTC or FMR)
            textstr.writeString(projfilenames[filecounter][1] + "\n"); // design matrix file (SDM)
        }
    } else {
        BrainVoyagerQX.PrintToLog("Kind of files to write to textfile unknown.");
        success = false;
    }
    filenamesfile.close();
    BrainVoyagerQX.PrintToLog("Wrote " + name + ".");
    return success;
}

scriptObj.readTextFileIntoProjectNamesArray = function(textfilename, kindoffiles) {

    var filecounter;    
    var success = true;
    BrainVoyagerQX.PrintToLog("Start reading: " + textfilename);
    var name = String(textfilename); // circumvention of problem with implicit strings
    var filenamesfile = new QFile(name);     
    filenamesfile.open(new QIODevice.OpenMode(QIODevice.ReadOnly));
    var textstr = new QTextStream(filenamesfile); 
    var nroffiles = parseInt(textstr.readLine());
    BrainVoyagerQX.PrintToLog("Number of projects: " + nroffiles);
    BrainVoyagerQX.PrintToLog("Kind of projects: " + ((kindoffiles == TEXTFILE_FOR_SSGLM) ? "single-subject GLM" : "unknown")); 

    if (kindoffiles == TEXTFILE_FOR_SSGLM) {

        projfilenames = new Array(nroffiles);
        if (success) {
            for (filecounter = 0; filecounter < nroffiles; filecounter++) {
                
                BrainVoyagerQX.PrintToLog("Files for GLM " + (filecounter+1) + ":");
                var groupnames = new Array(2);
                // *.vtc or *.fmr
                var filename = textstr.readLine();
                BrainVoyagerQX.PrintToLog(filename); 
                groupnames[0] = filename;
                // *.sdm
                var filename = textstr.readLine();
                BrainVoyagerQX.PrintToLog(filename); 
                groupnames[1] = filename;
                // save names in global array
                projfilenames[filecounter] = groupnames;                
            } 
        } 
        
        // set rest of parameters
        spbNrGLMs.value = nroffiles;
        if (nroffiles > 0) {
            var filetype = this.getFileType(projfilenames[0][0]);
            if (filetype == FILETYPE_VTC) {
                chbVTC.checked = true;
                chbMultiSubjGLM.enabled = true;
                chbRFXGLM.enabled = true;                
            } else if (filetype == FILETYPE_FMR) {
                chbFMR.checked = true;
                chbMultiSubjGLM.enabled = false;
                chbRFXGLM.enabled = false;
            } else {
                BrainVoyagerQX.PrintToLog("Filetype of functional file unknown.");
            }
        }
        
    } else {
        BrainVoyagerQX.PrintToLog("Kind of files unknown.");
    }
    filenamesfile.close();
    return success;
}

scriptObj.getPath = function(filename) { 

	var start = 0; // start searching from begin 
	var last = filename.lastIndexOf("/");
	var path = filename.substring(start, (last+1)); // now return path without filename
	return path;
}

scriptObj.noErrors = function() {

    var noerror;
    var errornr = BrainVoyagerQX.GetPluginIntParameter("ErrorNr");
    if (errornr != -1) {
        noerror = true;
    } else {
        var errormsg = BrainVoyagerQX.GetPluginStringParameter("ErrorMsg");
        BrainVoyagerQX.PrintToLog(errormsg);
        noerror = false;        
    }
    return noerror;
}  

/*scriptObj.getFileTypeFromArray = function(projfilenames, row, column) {

    var filetype = projfilenames[row][column].substring((projfilenames[row][column].length-3), projfilenames[row][column].length);
    return filetype;
}
*/
scriptObj.getFileType = function(filename) {


var filetype;
    var filetypestr = filename.substring((filename.length-3), filename.length);
    if (filetypestr == "vtc") {
        filetype = FILETYPE_VTC;
    } else if (filetypestr == "fmr") {
        filetype = FILETYPE_FMR;
    } else {
        filetype = FILETYPE_UNKNOWN;
    }
    return filetype;
}


//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

scriptObj.initDlg = function() {

    // find widgets
    spbNrGLMs               = BrainVoyagerQX.FindChild("spbNrGLMs");
    btnLoadFilenames        = BrainVoyagerQX.FindChild("btnLoadFilenames");
    lblVMRname              = BrainVoyagerQX.FindChild("lblVMRname");
    btnLoadVMR              = BrainVoyagerQX.FindChild("btnLoadVMR");
    rbtnReadTextfile        = BrainVoyagerQX.FindChild("rbtnReadTextfile");
    rbtnSelectViaDlg        = BrainVoyagerQX.FindChild("rbtnSelectViaDlg");
    chbCorrectSerCor        = BrainVoyagerQX.FindChild("chbCorrectSerCor");
    chbVTC                  = BrainVoyagerQX.FindChild("chbVTC");
    chbFMR                  = BrainVoyagerQX.FindChild("chbFMR");
    chbMultiSubjGLM         = BrainVoyagerQX.FindChild("chbMultiSubjGLM");
    chbRFXGLM               = BrainVoyagerQX.FindChild("chbRFXGLM");
    chbSeparateStudies      = BrainVoyagerQX.FindChild("chbSeparateStudies");
    chbSeparateSubjects     = BrainVoyagerQX.FindChild("chbSeparateSubjects");
    rbtnZtrfBaseline        = BrainVoyagerQX.FindChild("rbtnZtrfBaseline");
    rbtnZtrf                = BrainVoyagerQX.FindChild("rbtnZtrf");
    rbtnPSC                 = BrainVoyagerQX.FindChild("rbtnPSC");
    btnStart                = BrainVoyagerQX.FindChild("btnStart");

    // connect widgets to script functions   
    btnLoadVMR.clicked.connect(this, this.getAnatFile);
    btnLoadFilenames.clicked.connect(this, this.getDesMatAndFuncFiles);
    btnStart.clicked.connect(this, this.startProcessing);
 
}

returnScriptObj = function()
{
   return scriptObj; // it is important to return script object containing properties and functions to be called from BrainVoyager QX
}
returnScriptObj();
